#!/usr/bin/env python
#
# planck.py
# Reads the Planck Collaboration dust reddening maps.
#
# Copyright (C) 2016  Gregory M. Green
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import print_function, division

import os
import numpy as np
import healpy as hp
import astropy.io.fits as fits
import astropy.units as units

from .std_paths import *
from .healpix_map import HEALPixFITSQuery
from . import fetch_utils
from . import dustexceptions


class PlanckQuery(HEALPixFITSQuery):
    """
    Queries the Planck Collaboration (2013) dust map.
    """

    def __init__(self, map_fname=None, component='extragalactic'):
        """
        Args:
            map_fname (Optional[:obj:`str`]): Filename of the Planck map.
                Defaults to ```None``, meaning that the default location is
                used.
            component (Optional[:obj:`str`]): Which measure of reddening to use. There
                are seven valid components. Three denote reddening measures:
                ``'extragalactic'``, ``'tau'`` and ``'radiance'``. Four refer
                to dust properties: ``'temperature'``, ``'beta'``,
                ``'err_temp'`` and ``'err_beta'``. Defaults to
                ``'extragalactic'``.
        """

        if map_fname is None:
            map_fname = os.path.join(
                data_dir(),
                'planck',
                'HFI_CompMap_ThermalDustModel_2048_R1.20.fits'
            )

        if component.lower() in ('ebv','extragalactic'):
            field = 'EBV'
            self._scale = 1.
        elif component.lower() in ('tau','tau353','tau_353','optical depth'):
            field = 'TAU353'
            self._scale = 1.49e4
        elif component.lower() in ('radiance','r'):
            field = 'RADIANCE'
            self._scale = 5.4e5
        elif component.lower() in ('temperature','temp','t'):
            field = 'TEMP'
            self._scale = units.Kelvin
        elif component.lower() in ('sigma_temp','sigma_t','err_temp','err_t'):
            field = 'ERR_TEMP'
            self._scale = units.Kelvin
        elif component.lower() in ('beta','b'):
            field = 'BETA'
            self._scale = 1.
        elif component.lower() in ('sigma_beta','sigma_b','err_beta','err_b'):
            field = 'ERR_BETA'
            self._scale = 1.
        else:
            raise ValueError((
                "Invalid `component`: '{}'\n"
                "Valid components for reddening are 'extragalactic', 'tau', "
                "and 'radiance'. Valid components for dust properties are "
                "'temperature', 'err_temp', 'beta' and 'err_beta'."
                ).format(component))

        try:
            with fits.open(map_fname) as hdulist:
                super(PlanckQuery, self).__init__(
                    hdulist, 'galactic',
                    hdu='COMP-MAP',
                    field=field,
                    dtype='f4',
                    scale=self._scale
                )
        except IOError as error:
            print(dustexceptions.data_missing_message('planck',
                                                      'Planck Collaboration'))
            raise error

    def query(self, coords, **kwargs):
        """
        Returns E(B-V) (or a different Planck dust inference, depending on how
        the class was intialized) at the specified location(s) on the sky.

        Args:
            coords (:obj:`astropy.coordinates.SkyCoord`): The coordinates to
                query.

        Returns:
            A float array of the selected Planck component, at the given
            coordinates. The shape of the output is the same as the shape of the
            input coordinate array, ``coords``. If extragalactic E(B-V), tau_353
            or radiance was chosen, then the output has units of magnitudes of
            E(B-V). If the selected Planck component is temperature (or
            temperature error), then an :obj:`astropy.Quantity` is returned,
            with units of Kelvin. If beta (or beta error) was chosen, then the
            output is unitless.
        """
        return super(PlanckQuery, self).query(coords, **kwargs)


class PlanckGNILCQuery(HEALPixFITSQuery):
    """
    Queries the Planck Collaboration (2016) GNILC dust map.
    """
    def __init__(self, map_fname=None, load_errors=False):
        """
        Args:
            map_fname (Optional[:obj:`str`]): Filename of the Planck map.
                Defaults to ```None``, meaning that the default location is
                used.
            load_errors (Optional[:obj:`str`]): If ``True``, then the error
                estimates will be loaded as well, and returned with any query.
                If ``False`` (the default), then queries will only return the
                the reddening estimate, without any error estimate.
        """

        if load_errors:
            self._has_errors = True
            field = None
            dtype = [('EBV','f4'), ('EBV_err','f4')]
        else:
            self._has_errors = False
            field = 'TAU353'
            dtype = 'f4'

        if map_fname is None:
            map_fname = os.path.join(
                data_dir(),
                'planck',
                'COM_CompMap_Dust-GNILC-Model-Opacity_2048_R2.01.fits'
            )

        try:
            with fits.open(map_fname) as hdulist:
                super(PlanckGNILCQuery, self).__init__(
                    hdulist, 'galactic',
                    hdu=1,
                    field=field,
                    dtype=dtype,
                    scale=1.49e4
                )
        except IOError as error:
            print(dustexceptions.data_missing_message('planck',
                                                      'Planck GNILC'))
            raise error

    def has_errors(self):
        """
        Returns ``True`` if the error estimates have been loaded.
        """
        return self._has_errors

    def query(self, coords, **kwargs):
        """
        Returns E(B-V) at the specified location(s) on the sky.

        Args:
            coords (:obj:`astropy.coordinates.SkyCoord`): The coordinates to
                query.

        Returns:
            If the error estimates have been loaded, then a structured array
            containing ``'EBV'`` and ``'EBV_err'`` is returned. Otherwise,
            returns a float array of E(B-V), at the given coordinates. The
            shape of the output is the same as the shape of the input
            coordinate array, ``coords``.
        """
        return super(PlanckGNILCQuery, self).query(coords, **kwargs)


def fetch(which='2013'):
    """
    Downloads the Planck dust maps, placing them in the default ``dustmaps``
    directory. There are two different Planck dust maps that can be
    downloaded: the Planck Collaboration (2013) map and the "GNILC" (Planck
    Collaboration 2016) map.
    
    Args:
        which (Optional[:obj:`str`]): The name of the dust map to download.
            Should be either ``2013`` (the default) or ``GNILC``.
    """
    planck_maps = {
        '2013': {
            'url': 'http://pla.esac.esa.int/pla/aio/product-action?MAP.MAP_ID=HFI_CompMap_ThermalDustModel_2048_R1.20.fits',
            'md5': '8d804f4e64e709f476a63f0dfed1fd11',
            'fname': 'HFI_CompMap_ThermalDustModel_2048_R1.20.fits'
        },
        'GNILC': {
            'url': 'http://pla.esac.esa.int/pla/aio/product-action?MAP.MAP_ID=COM_CompMap_Dust-GNILC-Model-Opacity_2048_R2.01.fits',
            'md5': 'fc385c2ee5e82edf039cbca6e82d6872',
            'fname': 'COM_CompMap_Dust-GNILC-Model-Opacity_2048_R2.01.fits'
        }
    }
    if which not in planck_maps:
        raise ValueError(
            'Unknown map: "{}". Must be one of {}.'.format(
                which, tuple(planck_maps.keys())
            )
        )
    props = planck_maps[which]
    fname = os.path.join(data_dir(), 'planck', props['fname'])
    fetch_utils.download_and_verify(props['url'], props['md5'], fname=fname)


def main():
    from astropy.coordinates import SkyCoord
    q = PlanckQuery()
    c = SkyCoord([0., 180., 0.], [0., 0., 90.], frame='galactic', unit='deg')
    print(q(c))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# gaia_tge.py
# Reads the Gaia TGE dust reddening maps.
#
# Copyright (C) 2022  Gregory M. Green
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import print_function, division

import os
import numpy as np
import healpy as hp
from astropy.table import Table
import astropy.units as units

from .std_paths import *
from .healpix_map import HEALPixQuery
from . import fetch_utils
from . import dustexceptions


class GaiaTGEQuery(HEALPixQuery):
    """
    Queries the Gaia Total Galactic Extinction (Delchambre 2022) dust map,
    which contains estimates of monochromatic extinction, A0, in mags.
    """

    def __init__(self, map_fname=None, healpix_level='optimum'):
        """
        Args:
            map_fname (Optional[`str`]): Filename of the Gaia TGE map.
                Defaults to ``None``, meaning that the default location is
                used.
            healpix_level (Optional[`int` or `str`]): Which HEALPix
                level to load into the map. If "optimum" (the default), loads
                the optimum HEALPix level available at each location. If an
                `int`, instead loads the specified HEALPix level.
        """

        if map_fname is None:
            map_fname = os.path.join(
                data_dir(),
                'gaia_tge',
                'TotalGalacticExtinctionMap_001.csv.gz'
            )

        try:
            # Cannot use astropy ECSV reader, due to bug in processing
            # null values
            dtype = [
                ('solution_id', 'i8'),
                ('healpix_id', 'i8'),
                ('healpix_level', 'i1'),
                ('a0', 'f4'),
                ('a0_uncertainty', 'f4'),
                ('a0_min', 'f4'),
                ('a0_max', 'f4'),
                ('num_tracers_used', 'i4'),
                ('optimum_hpx_flag', '?'),
                ('status', 'i2')
            ]
            converters = {8: lambda x: x == '"True"'}
            d = np.genfromtxt(
                map_fname, comments='#', delimiter=',',
                encoding='utf-8', converters=converters,
                dtype=dtype
            )[1:]
        except IOError as error:
            print(dustexceptions.data_missing_message('gaia_tge',
                                                      'Gaia TGE'))
            raise error

        if isinstance(healpix_level, int):
            idx = (d['healpix_level'] == healpix_level)
            n_pix = np.count_nonzero(idx)
            if n_pix == 0:
                levels_avail = np.unique(d['healpix_level']).tolist()
                raise ValueError(
                    'Requested HEALPix level not stored in map. Available '
                    'levels: {}'.format(levels_avail)
                )
            hpx_sort_idx = np.argsort(d['healpix_id'][idx])
            idx = np.where(idx)[0]
            idx = idx[hpx_sort_idx]
        elif healpix_level == 'optimum':
            idx_opt = d['optimum_hpx_flag']
            # Upscale to highest HEALPix level
            hpx_level = d['healpix_level'][idx_opt]
            hpx_level_max = np.max(hpx_level)
            n_pix = 12 * 4**hpx_level_max
            # Index from original array to use in each pixel of final map
            idx = np.full(n_pix, -1, dtype='i8') # Empty pixel -> index=-1
            # Get the ring-ordered index of the optimal pixels
            idx_opt = np.where(idx_opt)[0]
            hpx_idx = d['healpix_id'][idx_opt]
            # Add pixels of each level to the map
            for level in np.unique(hpx_level):
                nside = 2**level
                idx_lvl = (hpx_level == level)
                # Get the nest-ordered index of optimal pixels at this level
                hpx_idx_nest = hpx_idx[idx_lvl]
                # Fill in index (in orig arr) of these pixels
                mult_factor = 4**(hpx_level_max-level)
                hpx_idx_base = hpx_idx_nest*mult_factor
                for offset in range(mult_factor):
                    idx[hpx_idx_base+offset] = idx_opt[idx_lvl]
        else:
            raise ValueError(
                '`healpix_level` must be either an integer or "optimum"'
            )

        bad_mask = (idx == -1)

        pix_val = d['a0'][idx]
        pix_val[bad_mask] = np.nan

        dtype = [
            ('a0_uncertainty', 'f4'),
            ('num_tracers_used', 'i4'),
            ('optimum_hpx_flag', 'bool')
        ]
        flags = np.empty(n_pix, dtype=dtype)
        for key,dt in dtype:
            flags[key] = d[key][idx]
            flags[key][bad_mask] = {'f4':np.nan, 'i4':-1, 'bool':False}[dt]

        super(GaiaTGEQuery, self).__init__(
            pix_val, True, 'icrs', flags=flags
        )

    def query(self, coords, **kwargs):
        """
        Returns a numpy array containing A0 at the specified
        location(s) on the sky. Optionally, returns a 2nd array containing
        flags at the same location(s).

        Args:
            coords (`astropy.coordinates.SkyCoord`): The coordinates to
                query.
            return_flags (Optional[`bool`]): If `True`, returns a 2nd array
                containing flags at each coordinate. Defaults to `False`.

        Returns:
            A numpy array containing A0 at the specified
            coordinates. The shape of the output is the same as the shape of
            the input coordinate array, ``coords``. If `return_flags` is
            `True`, a 2nd record array containing flags at each coordinate
            is also returned.
        """
        return super(GaiaTGEQuery, self).query(coords, **kwargs)


def fetch():
    """
    Downloads the Gaia Total Galactic Extinction (TGE) dust maps, placing
    it in the default ``dustmaps`` directory.
    """
    props = {
        'url': (
            'http://cdn.gea.esac.esa.int/Gaia/gdr3/Astrophysical_parameters/'
            'total_galactic_extinction_map/TotalGalacticExtinctionMap_001.csv.gz'
        ),
        'md5': '5f6271869b7e60960a955f08ca11dc37',
        'fname': 'TotalGalacticExtinctionMap_001.csv.gz'
    }
    fname = os.path.join(data_dir(), 'gaia_tge', props['fname'])
    fetch_utils.download_and_verify(props['url'], props['md5'], fname=fname)


def main():
    from astropy.coordinates import SkyCoord
    q = GaiaTGEQuery()
    c = SkyCoord([0., 180., 0.], [0., 0., 90.], frame='galactic', unit='deg')
    print(q(c))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# healpix_map.py
# A set of HEALPix map classes.
#
# Copyright (C) 2016  Gregory M. Green
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import print_function, division
import six

import numpy as np
import healpy as hp
import astropy.io.fits as fits

from .map_base import DustMap, coord2healpix


class HEALPixQuery(DustMap):
    """
    A class for querying HEALPix maps.
    """

    def __init__(self, pix_val, nest, coord_frame, flags=None):
        """
        Args:
            pix_val (array): Value of the map in every pixel. The length of the
                array must be of the form `12 * nside**2`, where `nside` is a
                power of two.
            nest (bool): `True` if the map uses nested ordering. `False` if
                ring ordering is used.
            coord_frame (str): The coordinate system that the HEALPix map is in.
                Should be one of the frames supported by `astropy.coordinates`.
        """
        self._nside = hp.pixelfunc.npix2nside(len(pix_val))
        self._pix_val = pix_val
        self._nest = nest
        self._frame = coord_frame
        self._flags = flags
        if (flags is not None) and (flags.shape[0] != pix_val.shape[0]):
            raise ValueError((
                'The shape of `flags` ({}) must match the shape '
                'of `pix_val` ({}) along the first axis.'
            ).format(flags.shape, pix_val.shape))
        super(HEALPixQuery, self).__init__()

    def query(self, coords, return_flags=False):
        """
        Args:
            coords (`astropy.coordinates.SkyCoord`): The coordinates to query.
            return_flags ([Optional[:obj:`bool`]): If `True`, return flags at
                each pixel. Only possible if flags were provided during
                initialization.

        Returns:
            A float array of the value of the map at the given coordinates. The
            shape of the output is the same as the shape of the coordinates
            stored by `coords`. If `return_flags` is `True`, then a second
            array, containing flags at each pixel, is also returned.
        """
        pix_idx = coord2healpix(coords, self._frame,
                                self._nside, nest=self._nest)
        sel_pix = self._pix_val[pix_idx]

        if return_flags:
            if self._flags is None:
                raise ValueError(
                    '`return_flags` is True, but the class was initialized '
                    'without flags.'
                )
            return sel_pix, self._flags[pix_idx]

        return self._pix_val[pix_idx]


class HEALPixFITSQuery(HEALPixQuery):
    """
    A HEALPix map class that is initialized from a FITS file.
    """

    def __init__(self, fname, coord_frame, hdu=0, field=None,
                                           dtype='f8', scale=None):
        """
        Args:
            fname (str, HDUList, TableHDU or BinTableHDU): The filename, HDUList
                or HDU from which the map should be loaded.
            coord_frame (str): The coordinate system in which the HEALPix map is
                defined. Must be a coordinate frame which ``astropy``
                understands.
            hdu (Optional[int or str]): Specifies which HDU to load the map
                from. Defaults to ``0``.
            field (Optional[int or str]): Specifies which field (column) to load
                the map from. Defaults to ``None``, meaning that ``hdu.data[:]``
                is used.
            dtype (Optional[str or type]): The data will be coerced to this
                datatype. Can be any type specification that numpy understands,
                including a structured datatype, if multiple fields are to be
                loaded. Defaults to ``'f8'``, for IEEE754 double precision.
            scale (Optional[:obj:`float`]): Scale factor to be multiplied into
                the data.
        """
        close_file = False

        if isinstance(fname, six.string_types):
            close_file = True
            hdulist = fits.open(fname)
            print(hdulist.info())
            hdu = hdulist[hdu]
        elif isinstance(fname, fits.HDUList):
            hdu = fname[hdu]
        elif (isinstance(fname, fits.TableHDU)
                or isinstance(fname, fits.BinTableHDU)):
            hdu = fname
        else:
            raise TypeError('`fname` must be a `str`, `HDUList`, `TableHDU` or '
                            '`BinTableHDU`.')

        if field is None:
            pix_val = np.array(hdu.data[:].ravel().astype(dtype))
        else:
            pix_val = np.array(hdu.data[field][:].ravel().astype(dtype))

        if scale is not None:
            names = pix_val.dtype.names
            if names is None:
                pix_val *= scale
            else:
                for n in names:
                    pix_val[n] *= scale

        nest = hdu.header.get('ORDERING', 'NESTED').strip() == 'NESTED'

        if close_file:
            hdulist.close()

        super(HEALPixFITSQuery, self).__init__(pix_val, nest, coord_frame)
